#include <ESP8266WiFi.h>
extern "C" {
#include "user_interface.h"
}

//NodeMCU WiFi Data Logger - http://pvmonitor.pl (2016-06-28)
//Łukasz Janikowski (c)2016
//http://elektronika.5v.pl

//Board: NodeMCU 1.0 (ESP-12E Module)
//CPU Frequency: 80MHz
//Flash Size: 4M (3M SPIFFS)
//Upload Speed: 115200


//parametry ustawiane przez użytkownika
//##############################################################################################################################################################
#define mySSID  "NASZEWIFI"       //nazwa domowej sieci WiFi
#define myPASS  "naszehaslo"      //hasło do domowej sieci WiFi (jeżeli sieć jest niezabezpieczona - "")

#define pvSERV  "pvmonitor.pl"  //adres serwera na który wysyłamy dane, np. pvmonitor.pl lub 37.187.66.191
#define pvID    "XXX"           //nazwa użytkownika - konto w serwisie http://pvmonitor.pl
#define pvPASS  "xxxyyyzzz"    //hasło do konta w serwisie http://pvmonitor.pl

#define WYSYLAJ_PRODUKCJE   1   //0 = pomiń, 1 = wysyłaj wartość produkcji energii na serwer
#define WYSYLAJ_ZUZYCIE     1   //0 = pomiń, 1 = wysyłaj wartość zużycia energii na serwer
#define WYSWIETL_SIECI      0   //0 = wyłącz, 1 = wyszukaj i wyświetl dostępne sieci WiFi (akcja wywoływana co 1 minutę) 

#define wifiSTATIC          0   //0 = pozyskaj adres IP automatycznie z DHCP routera, 1 = ustaw stałe IP
IPAddress wifiIP(192, 168, 0, 44);    //stały adres IP
IPAddress wifiMASK(255, 255, 255, 0); //maska podsieci
IPAddress wifiGATE(192, 168, 0, 1);   //brama domyślna
IPAddress wifiDNS(192, 168, 0, 1);    //adres serwera DNS (routera)
//##############################################################################################################################################################


//ustawienia sprzętowe
#define IMP1 D1     //definicje pinów do zliczania impulsów z liczników energii
#define IMP2 D2
#define IMP3 D3
#define IMP4 D5
#define IMP5 D6
#define IMP6 D7

#define LED 2       //pin wbudowanej diody LED
#define PWM_LOW 250 //jasności diody LED
#define PWM_MID 200
#define PWM_HIGH 0


//zmienne globalne 
unsigned long produkcja_Wh;
unsigned long zuzycie_Wh;
byte timer_60s; 
bool tick_1s;
os_timer_t myTimer;
bool wifi_status;
byte send_data;
byte search_wifi;
WiFiClient client;

//stałe tekstowe
const char* http = "http://";

//funkcje
//##############################################################################################################################################################
void timerCallback(void *pArg) {
  tick_1s = true;
}

void timer_init(void) {
  os_timer_setfn(&myTimer, timerCallback, NULL);
  os_timer_arm(&myTimer, 1000, true);
}

void licz_produkcje(void) {
  produkcja_Wh++;
  if(produkcja_Wh>999999999) produkcja_Wh = 0;
}

void licz_zuzycie(void) {
  zuzycie_Wh++;
  if(zuzycie_Wh>999999999) zuzycie_Wh = 0;
}

void connect_WiFi(void) {
  Serial.print(F("Connecting to "));
  Serial.print(mySSID);
  WiFi.mode(WIFI_STA);
  WiFi.begin(mySSID, myPASS);
#if wifiSTATIC == 1
  WiFi.config(wifiIP, wifiGATE, wifiMASK, wifiDNS);
#endif 
  byte count = 0;
  while (WiFi.status() != WL_CONNECTED) {
    delay(100);
    Serial.print(F("."));
    count++;
    if(count==100) break;
  }
  Serial.println();
  if(count==100) {
    Serial.println(F("ERROR: Not connected to WiFi!"));
    wifi_status = false;
  } else {
    Serial.print(F("Connected - IP: "));
    Serial.print(WiFi.localIP());
    Serial.print(F("  RSSI: "));
    Serial.print(WiFi.RSSI());
    Serial.println(F(" dBm"));
    analogWrite(LED, PWM_LOW);
    wifi_status = true;
  }
  Serial.println();
}

void check_and_reconnect_wifi(void) {
  if(WiFi.status() != WL_CONNECTED) {
    analogWrite(LED, 255);
    if(wifi_status) {
      Serial.println(F("WiFi disconnected!"));
      Serial.println();
    }
    wifi_status = false;
    connect_WiFi();
  }
}


void send_data_to_server(void) {  
  analogWrite(LED, PWM_HIGH);
  if(client.connect(pvSERV, 80)) {
    Serial.print("Sending data...   ");
    client.print("GET ");
    client.print(http);
    client.print(pvSERV);
    client.print("/pv/get.php?idl=");
    client.print(pvID);
    client.print("&p=");
    client.print(pvPASS);
#if WYSYLAJ_PRODUKCJE == 1
      client.print("&f1=");
      client.print(produkcja_Wh/1000);
      client.print(".");
      client.print(produkcja_Wh%1000/100);
      client.print(produkcja_Wh%100/10);
      client.print(produkcja_Wh%10);
#endif
#if WYSYLAJ_ZUZYCIE == 1 
      client.print("&f4=");
      client.print(zuzycie_Wh/1000);
      client.print(".");
      client.print(zuzycie_Wh%1000/100);
      client.print(zuzycie_Wh%100/10);
      client.print(zuzycie_Wh%10);
#endif
    client.print(" HTTP/1.0\r\n\r\n");
  }
  String response = client.readStringUntil('\r');
  String date = client.readStringUntil('\r');
  date.remove(0, 1);
  if(response.startsWith("HTTP/1.1 200 OK")) {
    Serial.print(F("OK   "));
#if WYSYLAJ_PRODUKCJE == 1
    Serial.print(F("P: "));    
    Serial.print(produkcja_Wh/1000);
    Serial.print(F(","));
    Serial.print(produkcja_Wh%1000/100);
    Serial.print(produkcja_Wh%100/10);
    Serial.print(produkcja_Wh%10);
    Serial.print(F(" kWh"));
#endif
#if WYSYLAJ_PRODUKCJE == 1 && WYSYLAJ_ZUZYCIE == 1
    Serial.print(F(" / "));
#endif
#if WYSYLAJ_ZUZYCIE == 1 
    Serial.print(F("U: ")); 
    Serial.print(zuzycie_Wh/1000);
    Serial.print(F(","));
    Serial.print(zuzycie_Wh%1000/100);
    Serial.print(zuzycie_Wh%100/10);
    Serial.print(zuzycie_Wh%10);
    Serial.print(F(" kWh   "));
#endif    
  Serial.println(date);
  send_data = 0;
  } else {
    send_data++;
    if(send_data>3) {
      send_data = 0;
      Serial.println(F("ERROR"));
    } else {
      Serial.println(F("Repeating..."));
    }
  }
  client.flush();
  client.stop();
  analogWrite(LED, PWM_LOW);
}


void print_encryption_type(int thisType) {
   switch(thisType) {
     case ENC_TYPE_WEP:
       Serial.print(F(" WEP"));
       break;
     case ENC_TYPE_TKIP:
       Serial.print(F(" WPA"));
       break;
     case ENC_TYPE_CCMP:
       Serial.print(F("WPA2"));
       break;
     case ENC_TYPE_NONE:
       Serial.print(F("None"));
       break;
     case ENC_TYPE_AUTO:
       Serial.print(F("Auto"));
       break;
   }
}

void print_networks_list(void) {
   int numSsid = WiFi.scanNetworks();
   Serial.println();
   if (numSsid == -1) {
     Serial.println(F("WiFi networks not found!"));
   } else {
     Serial.println(F("Available WiFi networks:"));

    for (int thisNet = 0; thisNet < numSsid; thisNet++) {
     Serial.print(thisNet+1);
     Serial.print(F(". "));
     Serial.print(WiFi.RSSI(thisNet));
     Serial.print(F(" dBm, "));
     print_encryption_type(WiFi.encryptionType(thisNet));
     Serial.print(F(", "));
     Serial.println(WiFi.SSID(thisNet));
    }
   }
   Serial.println();
}


//start programu
//##############################################################################################################################################################
void setup() {
  pinMode(LED, OUTPUT);           //wyjście dla wbudowanej diody LED
  analogWriteFreq(200);           //częstotliwość PWM LED
  analogWriteRange(255);          //rozdzielczość PWM LED
  analogWrite(LED, 255);          //wyłączenie diody LED
  
  pinMode(IMP1, INPUT_PULLUP);    //wejście impulsowe IMP1
  attachInterrupt(IMP1, licz_produkcje, FALLING);   //przerwanie dla wejścia IMP1 wyzwalane zboczem opadającym  
  pinMode(IMP2, INPUT_PULLUP);    //wejście impulsowe IMP2
  attachInterrupt(IMP2, licz_produkcje, FALLING);   //przerwanie dla wejścia IMP2 wyzwalane zboczem opadającym  
  pinMode(IMP3, INPUT_PULLUP);    //wejście impulsowe IMP3
  attachInterrupt(IMP3, licz_produkcje, FALLING);   //przerwanie dla wejścia IMP3 wyzwalane zboczem opadającym  
  
  pinMode(IMP4, INPUT_PULLUP);    //wejście impulsowe IMP4
  attachInterrupt(IMP4, licz_zuzycie, FALLING);     //przerwanie dla wejścia IMP4 wyzwalane zboczem opadającym
  pinMode(IMP5, INPUT_PULLUP);    //wejście impulsowe IMP5
  attachInterrupt(IMP5, licz_zuzycie, FALLING);     //przerwanie dla wejścia IMP5 wyzwalane zboczem opadającym
  pinMode(IMP6, INPUT_PULLUP);    //wejście impulsowe IMP6
  attachInterrupt(IMP6, licz_zuzycie, FALLING);     //przerwanie dla wejścia IMP6 wyzwalane zboczem opadającym
  produkcja_Wh = 0;
  zuzycie_Wh = 0;

  Serial.begin(115200);           //inicjalizacja portu szeregowego
  Serial.println();
  Serial.println(F("NodeMCU WiFi Data Logger - http://pvmonitor.pl"));
  Serial.println(F("Created by Lukasz Janikowski  http://elektronika.5v.pl"));

  connect_WiFi();                 //podłączenie do sieci WLAN
  timer_init();                   //inicjalizacja timera programowego
}

//pętla główna
//##############################################################################################################################################################
void loop() {

  if(tick_1s){
    check_and_reconnect_wifi();
    timer_60s++;   
    if(timer_60s == 60) {
      send_data = 1;
      search_wifi = 1;
      timer_60s = 0;
    }
    if(send_data && wifi_status) {
      send_data_to_server();
    }
#if WYSWIETL_SIECI == 1
    if(search_wifi) {
      print_networks_list();
      search_wifi = 0;
    }
#endif
    tick_1s = false;
  }
  
}
